#include "burp.h"
#include "prog.h"

static unsigned long long totalcounter=0;
static unsigned long long filecounter=0;
static unsigned long long changedcounter=0;
static unsigned long long unchangedcounter=0;
static unsigned long long newcounter=0;
static unsigned long long directorycounter=0;
static unsigned long long specialcounter=0;
static unsigned long long hardlinkcounter=0;
static unsigned long long softlinkcounter=0;
static unsigned long long warningcounter=0;
static unsigned long long bytecounter=0;
static unsigned long long recvbytecounter=0;
static unsigned long long sentbytecounter=0;
static unsigned long long encryptedcounter=0;

void reset_filecounter(void)
{
	totalcounter=0;
	filecounter=0;
	changedcounter=0;
	unchangedcounter=0;
	newcounter=0;
	directorycounter=0;
	specialcounter=0;
	hardlinkcounter=0;
	softlinkcounter=0;
	warningcounter=0;
	bytecounter=0;
	recvbytecounter=0;
	sentbytecounter=0;
	encryptedcounter=0;
}

static const char *bytes_to_human(unsigned long long counter)
{
	static char ret[32]="";
	float div=(float)counter;
	char units[3]="";

	if(div<1024) return "";

	if((div/=1024)<1024)
		snprintf(units, sizeof(units), "KB");
	else if((div/=1024)<1024)
		snprintf(units, sizeof(units), "MB");
	else if((div/=1024)<1024)
		snprintf(units, sizeof(units), "GB");
	else if((div/=1024)<1024)
		snprintf(units, sizeof(units), "TB");
	else if((div/=1024)<1024)
		snprintf(units, sizeof(units), "EB");
	else
	{
		div/=1024;
		snprintf(units, sizeof(units), "PB");
	}
	snprintf(ret, sizeof(ret), " (%.2f %s)", div, units);
	//strcat(ret, units);
	//strcat(ret, ")");
	return ret;
}

static void pcounter(const char *str, unsigned long counter)
{
	if(counter) logp(str, counter);
}

static void bcounter(const char *str, unsigned long long counter)
{
	if(counter)
	{
		char msg[128]="";
		snprintf(msg, sizeof(msg),
			str, counter, bytes_to_human(counter));
		logp("%s", msg);
	}
}

void end_filecounter(int print, enum action act)
{
	if(print) printf(" %llu\n\n", totalcounter);

	if(act==ACTION_RESTORE)
	{
		logp("Restored files:            %llu\n", filecounter);
		pcounter("Restored encrypted files:  %llu\n", encryptedcounter);
	}
	else if(act==ACTION_VERIFY)
	{
		logp("Verified files:            %llu\n", filecounter);
		pcounter("Verified encrypted files:  %llu\n", encryptedcounter);
	}
	else
	{
		pcounter("Files:                     %llu\n", filecounter);
		pcounter("Encrypted files:           %llu\n", encryptedcounter);
		pcounter("New files:                 %llu\n", newcounter);
		pcounter("Changed files:             %llu\n", changedcounter);
		pcounter("Unchanged files:           %llu\n", unchangedcounter);
	}
	pcounter("Directories:               %llu\n", directorycounter);
	pcounter("Soft links:                %llu\n", softlinkcounter);
	pcounter("Hard links:                %llu\n", hardlinkcounter);
	pcounter("Special files:             %llu\n", specialcounter);
	//logp("\n");
	pcounter("Total:                     %llu\n", totalcounter);
	//logp("\n");
	pcounter("Warnings:                  %llu\n", warningcounter);
	bcounter("Bytes received:  %llu%s\n", recvbytecounter);
	bcounter("Bytes sent:      %llu%s\n", sentbytecounter);
	bcounter("Bytes in backup: %llu%s\n", bytecounter);
	//logp("\n");
	reset_filecounter();
}

void do_filecounter(char c, int print)
{
	if(print)
	{
		if(!totalcounter && !warningcounter) printf("\n");
		printf("%c", c);
	}
	switch(c)
	{
		case 'f': ++filecounter; break;
		case 'x': ++changedcounter; break;
		case 'F': ++newcounter; break;
		case 'd': ++directorycounter; break;
		case 's': ++specialcounter; break;
		case 'l': ++softlinkcounter; break;
		case 'L': ++hardlinkcounter; break;
		case 'w': ++warningcounter; return; // do not add to total
		case 'y': ++encryptedcounter; break;
	}
	if(!(++totalcounter%64) && print) printf(" %llu\n", totalcounter);
	fflush(stdout);
}

void do_filecounter_bytes(unsigned long long bytes)
{
	bytecounter+=bytes;
}

void do_filecounter_sentbytes(unsigned long long bytes)
{
	sentbytecounter+=bytes;
}

void do_filecounter_recvbytes(unsigned long long bytes)
{
	recvbytecounter+=bytes;
}
